package fr.pandacube.lib.paper.util;

import org.bukkit.entity.Player;

/**
 * A utility for managing Player experience properly.<br/>
 * <a href="https://gist.github.com/Jikoo/30ec040443a4701b8980">Github Gist - Jikoo</a>
 *
 * @author Jikoo
 */
public class ExperienceUtil {

	/**
	 * Calculates a player's total exp based on level and progress to next.
	 *
	 * @see <a href="http://minecraft.gamepedia.com/Experience#Leveling_up">Experience (#leveling up) - Minecraft Wiki</a>
	 *
	 * @param player the Player
	 *
	 * @return the amount of exp the Player has
	 */
	public static int getExp(Player player) {
		return getExpFromLevel(player.getLevel()) + Math.round(getExpToNext(player.getLevel()) * player.getExp());
	}

	/**
	 * Calculates total experience based on level.
	 *
	 * @see <a href="http://minecraft.gamepedia.com/Experience#Leveling_up">Experience (#leveling up) - Minecraft Wiki</a>
	 *
	 * @param level the level
	 *
	 * @return the total experience calculated
	 */
	public static int getExpFromLevel(int level) {
		if (level > 30) return (int) (4.5 * level * level - 162.5 * level + 2220);
		if (level > 15) return (int) (2.5 * level * level - 40.5 * level + 360);
		return level * level + 6 * level;
	}

	/**
	 * Calculates level based on total experience.
	 *
	 * @param exp the total experience
	 *
	 * @return the level calculated
	 */
	public static double getLevelFromExp(int exp) {
		if (exp > 1395) return (Math.sqrt(72 * exp - 54215) + 325) / 18;
		if (exp > 315) return Math.sqrt(40 * exp - 7839) / 10 + 8.1;
		if (exp > 0) return Math.sqrt(exp + 9) - 3;
		return 0;
	}

	/**
	 * @see <a href="http://minecraft.gamepedia.com/Experience#Leveling_up">Experience (#leveling up) - Minecraft Wiki</a>
	 *
	 *      "The formulas for figuring out how many experience orbs you need to
	 *      get to the next level are as follows:
	 *      Experience Required = 2[Current Level] + 7 (at levels 0-15)
	 *      5[Current Level] - 38 (at levels 16-30)
	 *      9[Current Level] - 158 (at level 31+)"
	 */
	private static int getExpToNext(int level) {
		if (level > 30) return 9 * level - 158;
		if (level > 15) return 5 * level - 38;
		return 2 * level + 7;
	}

	/**
	 * Change a Player's exp.
	 * <p>
	 * This method should be used in place of {@link Player#giveExp(int)}, which
	 * does not properly
	 * account for different levels requiring different amounts of experience.
	 *
	 * @param player the Player affected
	 * @param exp the amount of experience to add or remove
	 */
	public static void changeExp(Player player, int exp) {
		exp += getExp(player);

		if (exp < 0) exp = 0;

		double levelAndExp = getLevelFromExp(exp);

		int level = (int) levelAndExp;
		player.setLevel(level);
		player.setExp((float) (levelAndExp - level));
	}

}
